

/* asn1_helper.c
*
* Routine to augment previously existing ASN1 routines
*
* $Id: asn1_helper.c,v 1.0 2004/012/12  Exp $
*
* Herbert Falk <herb@sisconet.com>
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>

#include <stdio.h>
#include <string.h>

#include "asn1.h"
#include "packet-frame.h"
#include "prefs.h"

#include <epan/strutil.h>
#include "asn1_helper.h"
#include "math.h"


#define SECONDS_PER_DAY		(60*60*24)
#define DAYS_FROM_1970_TO_1984	( (11 *365) + (3 * 366))  /*takes into account leap years 1972, 1976, 1982 */
#define MAX_MSEC_PER_DAY (1000 * SECONDS_PER_DAY)


/* prefs from proto-iso9506.c for showing ASN.1 as subtrees or printf */
static gboolean display_nodes_as_subtrees;


static char unknown_string[100];


/* Add a text-only node to the proto_tree */
static proto_item *
asn1h_proto_tree_add_text(proto_tree *tree, int hf, tvbuff_t *tvb, gint start, gint length, const char *s)
{
	proto_item *ti;

	if (!s || !strlen(s)) 
		return(0);

	if (hf < 0)
	{
		ti = proto_tree_add_text (tree, tvb, start, length, "%s", s);
	}
	else
	{
		header_field_info *hfinfo = proto_registrar_get_nth(hf);
		if (hfinfo)
		{
			switch (hfinfo->type)
			{
			case FT_NONE:
				ti = proto_tree_add_none_format (tree, hf, tvb, start, length, "%s", s);
				break;
			case FT_STRING:
			case FT_STRINGZ:
			{
				ti = proto_tree_add_string(tree, hf, tvb, start, length, s ? s : "");
				break;
			}	
			default:
				if (s && !display_nodes_as_subtrees) 
					proto_tree_add_text(tree, tvb, start, length, "%s", s);				
				ti = proto_tree_add_uint(tree, hf, tvb, start, length, 0);
				break;
			}
		}
		else
			ti = proto_tree_add_text (tree, tvb, start, length, "%s", s);

	}
	return(ti);
}

/* Add a text and unsigned interger node to the proto_tree */
static proto_item *
asn1h_proto_tree_add_text_uint(proto_tree *tree, int hf, tvbuff_t *tvb, gint start, gint length, const char *s, guint32 myuint)
{
	proto_item *ti;

	if (hf < 0)
	{
		ti = proto_tree_add_text (tree, tvb, start, length, "%s %u", s, myuint);
	}
	else
	{
		header_field_info *hfinfo = proto_registrar_get_nth(hf);
		if (hfinfo)
		{
			switch (hfinfo->type)
			{
			case FT_NONE:
				ti = proto_tree_add_none_format (tree, hf, tvb, start, length, "%s %u", s ? s : "", myuint);
				break;
			case FT_STRING:
			case FT_STRINGZ:
			{
				ti = proto_tree_add_string(tree, hf, tvb, start, length,s ? s : "");			
				break;
			}	
			default:
#if 0
				if (s && !display_nodes_as_subtrees)
					proto_tree_add_text(tree, tvb, start, length, "%s", s);				
				ti = proto_tree_add_uint(tree, hf, tvb, start, length, myuint);
				if (s && !display_nodes_as_subtrees) 
#endif
				if (s) 
				  ti = proto_tree_add_uint_format_value(tree,hf,tvb,start,length,myuint,"%s %u",s,myuint);
				else
				  ti = proto_tree_add_uint(tree, hf, tvb, start, length, 0);
				break;



			}
		}
		else
			ti = proto_tree_add_text (tree, tvb, start, length, "%s %u", s, myuint);

	}
	return(ti);
}


/* Add a text and signed interger node to the proto_tree */
static proto_item *
asn1h_proto_tree_add_text_int(proto_tree *tree, int hf, tvbuff_t *tvb, gint start, gint length, const char *s, gint32 myint)
{
	proto_item *ti;

	if (hf < 0)
	{
		ti = proto_tree_add_text (tree, tvb, start, length, "%s %d", s, myint);
	}
	else
	{
		header_field_info *hfinfo = proto_registrar_get_nth(hf);
		if (hfinfo)
		{
			switch (hfinfo->type)
			{
			case FT_NONE:
				ti = proto_tree_add_none_format (tree, hf, tvb, start, length, "%s %d", s ? s : "", myint);
				break;
			case FT_STRING:
			case FT_STRINGZ:
			{
				ti = proto_tree_add_string(tree, hf, tvb, start, length,s ? s : "");			
				break;
			}	
			default:
				if (s && !display_nodes_as_subtrees) 
					proto_tree_add_text(tree, tvb, start, length, "%s", s);				
				ti = proto_tree_add_int(tree, hf, tvb, start, length, myint);
				break;
			}
		}
		else
			ti = proto_tree_add_text (tree, tvb, start, length, "%s %d", s, myint);

	}
	return(ti);
}

static void
find_enum_string_value(guint myint,ENUM_XLAT *xlat,char **ret_ptr)
{
guint num_entries;
guint i;
ENUM_LOOKUP *lookup;

  sprintf(unknown_string,"unknown value(%d)",myint);
  *ret_ptr = unknown_string;
  if( xlat!=NULL)
  {
   num_entries = xlat->num_entries;
   lookup = xlat->lookup;
   for (i=0;i<num_entries; ++i)
   {
     if (lookup->value == myint)
	    {
	    sprintf(unknown_string,"%s (%d)",lookup->stringvalue, myint);
		break;
		}
	else
	   ++lookup;

   }
  }
}


int
asn1h_tag_get(ASN1_SCK *asn1, guint *tag)
{
    int    ret;
    guchar ch;
	guint int_tag = 0;
	guint tag1=0;

	   ret = asn1_octet_decode(asn1, &ch);
	   if (ret == ASN1_ERR_NOERROR)			/*then get the information  */
	    {
	    int_tag = (ch & 0xe0 )<<24;
	    if ( (ch & 0x1f) == 0x1f)				/*then the next byte contains the tag value */
	      {
		  do {
		     ret = asn1_octet_decode(asn1, &ch );	/*go get the next part of the tag value		*/
		     if (ret!=ASN1_ERR_NOERROR)
		       return ret;
		     tag1<<=7;
		     tag1 |= ch & 0x7f;
		     }while ((ch & 0x80) == 0x80);		/*keep going if the tag extension bit is set */
		  int_tag = int_tag + tag1;
		  *tag = int_tag;
		  }
	    else									/*there is no tag extension */
	      {
		  int_tag += (guint)(ch & 0x1f);
		  *tag = int_tag;
		  }
        return ASN1_ERR_NOERROR;
		}
	   else
	    {
		*tag = 0xffffffff;
	     return ASN1_ERR_WRONG_TYPE;
		}
}


/*function to calculate the actual indefinite length		*/
static void
calc_indef_length (ASN1_SCK *asn1,guint *indef_len)
{
ASN1_SCK asn1_indef;
guint tag;
guint indef_level = 1;
guint internal_length;
guint initial_offset = asn1->offset;
gboolean def;
guchar *myoctets;



	memcpy (&asn1_indef, asn1, sizeof(ASN1_SCK));			/*copy and save the information */

	/*now need to go looking for the end	*/

    while(indef_level>0)
	{
      asn1h_tag_get(&asn1_indef, &tag);
      asn1_length_decode (&asn1_indef, &def, &internal_length);
      if(def==FALSE)
	    ++indef_level;
	  else
	   {		/*we have a definite length and need to check if it is the end of indef	*/
	   if ( (tag == 0x0)  && (internal_length == 0x0) )	/*then we have reached the end	*/
	     --indef_level;
	   else
	     {
		 /*need to skip a number of octets					*/
   	     asn1_string_value_decode (&asn1_indef, internal_length, &myoctets);
		 g_free(myoctets);
         }
      }
  	}
	*indef_len = asn1_indef.offset - initial_offset -2;
}


/*
 * NAME:        asn1_length_decode
 * SYNOPSIS:    int asn1_length_decode
 *                  (
 *                      ASN1_SCK *asn1,
 *                      gboolean *def,
 *                      guint    *len
 *                  )
 * DESCRIPTION: Decodes an ASN1 length.
 *              Parameters:
 *              asn1: pointer to ASN1 socket.
 *              def: Boolean - TRUE if length definite, FALSE if not
 *              len: length, if length is definite
 * DESCRIPTION: Decodes a definite or indefinite length.
 * RETURNS:     ASN1_ERR value (ASN1_ERR_NOERROR on success)
 */
int
asn1h_length_decode(ASN1_SCK *asn1, gboolean *def, guint *len)
{
    int    ret;
    guchar ch, cnt;
    guint 	indef_len=0;

    ret = asn1_octet_decode (asn1, &ch);
    if (ret != ASN1_ERR_NOERROR)
        return ret;
    if (ch == 0x80)
        *def = FALSE;		/* indefinite length */
    else {
        *def = TRUE;		/* definite length */
        if (ch < 0x80)
            *len = ch;
        else {
            cnt = (guchar) (ch & 0x7F);
            *len = 0;
            while (cnt > 0) {
                ret = asn1_octet_decode (asn1, &ch);
                if (ret != ASN1_ERR_NOERROR)
                    return ret;
                *len <<= 8;
                *len |= ch;
                cnt--;
            }
        }
    }

	/*check if indefinite length, if so, then find the actual length 				*/
	/* note that the length will be 2 octets less than the actual number of octets	*/
	/*this is so that recursive structures can be handled propely					*/
  if (*def == FALSE)
   {
	calc_indef_length (asn1,&indef_len);
	*len = indef_len - 2;
   }

	return ASN1_ERR_NOERROR;
}


/******************************************************************
 The following routines are used to do MMS UTC TIME Conversion	*/
static gdouble
convert_fractionsOfSeconds (guint32 initialvalue)
{
  guint32 maskedvalue;
  int i;
  gdouble init = 0.5;
  gdouble outputVal = 0.0;

  maskedvalue = initialvalue & 0xffffff00;		/*mask off the last byte (time quality) */
  for (i = 0; i<24;++i)
  {
	 if (maskedvalue & 0x80000000)
		 outputVal += init;

	 maskedvalue  = maskedvalue <<1;
	 init = init/2;
  }
  return(outputVal);
}

/****************************************************************************************/
/* function to handle MMS_FLOATS														*/
static void
convert_to_MMSFLOAT (ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guint32 total_length=*total_len;
guint32 exponent_widthr;
guint32 exponent_width;
guint32 exponent_value = 0;
gdouble exponent;

guint32 fraction_int=0;
guchar *mms_float_data;
gdouble sign=1.0;
guchar mask = 0x40;
guint32 i;
guint32 byte_offset=0;
guchar data_to_be_masked;
guint32 fraction_num_bits;
guint32 saved_length;
double double_value = 0.0;
guchar mydisplaystring[128];
gdouble fraction_adder = 0.5;
gdouble fraction= 0.0;

 /*get the length  */
 saved_length = (asn1->offset)-init_offset + total_length;

 if (total_length >= 2)			/*then we have a value */
 {
    asn1_uint32_value_decode (asn1, 1, &exponent_widthr);
	exponent_width = exponent_widthr & 0xff;			/*make sure to get the correct byte*/

	--total_length;
    fraction_num_bits = (total_length * 8) -exponent_width -1;
	asn1_string_value_decode (asn1, total_length, &mms_float_data);/*allocate a buffer for the worst case size*/

/*then we have a negative sign bit */

	if ((*mms_float_data) & 0x80)  
	{
		sign = -1.0;
	}

	/*now get the exponent */
	for (i=0;i<exponent_width; ++i)
	{
      data_to_be_masked = *(mms_float_data+byte_offset);
	  exponent_value = exponent_value * 2;
	  if (data_to_be_masked & mask)  /*then have a bit */
	  {
		  ++exponent_value;
	  }
	  mask = mask >> 1;
	  if (mask == 0x0)
	  {
		  mask = 0x80;
		  ++byte_offset;
	  }

	}


	for(i=0;i<fraction_num_bits;++i)
	{
      data_to_be_masked = *(mms_float_data+byte_offset);
	  if (data_to_be_masked & mask)  /*then have a bit */
	  {
		  fraction+=fraction_adder;
	  }
	  mask = mask >> 1;
	  fraction_adder = fraction_adder/2;
	  if (mask == 0x0)
	  {
		  mask = 0x80;
		  ++byte_offset;
	  }
	}


 exponent =  (gdouble)exponent_value -  (gdouble)(ldexp (1,(exponent_width - 1) ))  +1.0;
/* double_value = sign * (gdouble)(( 1.0+  (gdouble)(1.0/fraction_int))* ldexp(1,(gint32)exponent));
 */


 double_value = sign * (gdouble)( 1.0+  fraction)* (gdouble)(ldexp(1,(gint32)exponent));




 sprintf(mydisplaystring,"%s%s %f ",display_string,label_text, double_value);
 asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_length , mydisplaystring);

 g_free(mms_float_data);
 }

}



static void
convert_to_MMSUTCTIME(ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guint32 total_length=*total_len;
time_t  ts;
guint32 intstorage;
guchar *myptr;
struct tm *t;
gdouble fractions;
guchar display_text[100];
gboolean need_offset= FALSE;

	ts = tvb_get_ntohl(asn1->tvb, asn1->offset);
	if (ts & 0x80000000)
	  {	   
	  /* need to go beyond 2038 */
	  need_offset = TRUE;
	  ts -= 0x7fe81780;
	  }

	intstorage = tvb_get_ntohl(asn1->tvb, asn1->offset+4);
	fractions = convert_fractionsOfSeconds (intstorage);
	asn1_string_value_decode (asn1, total_length, &myptr);
	g_free(myptr);
	t = gmtime(&ts);
	fractions += t->tm_sec;
	if (need_offset == FALSE)
 	  sprintf(display_text,"%s%s %04d-%02d-%02d %02d:%02d.%f  Timequality: %02x",
					display_string, label_text, 
					1900+t->tm_year,t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min,
                     fractions, (guchar)intstorage & 0xff);
	else
 	  sprintf(display_text,"%s%s %04d-%02d-%02d %02d:%02d.%f  Timequality: %02x",
					display_string, label_text, 
					2038+t->tm_year,t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min,
                     fractions, (guchar)intstorage & 0xff);

	asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, (guint)asn1->offset - init_offset,display_text);

}

/****************************************************************************************/
/* function to handle ASN.1 GENERALIZED Time										*/
int convert_to_GENERALTIME (ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guint32 total_length=*total_len;
guchar *myptr;
guchar *display_text;
guchar year[5];
guchar month[3];
guchar day[3];
guchar hour[3];
guchar minutes[3];
guchar *seconds;

	display_text = g_malloc(total_length+40+strlen(display_string)+strlen(label_text));

	asn1_string_value_decode (asn1, total_length, &myptr);
    /*need to get the data so it can be printed*/
	if(total_length< 15) /*then we have an invalid GTIME */
 	  sprintf(display_text,"%s%s Invalid GTIME (%s)",display_string, label_text , myptr);
	else
	  {  /*we can decode the time */
	  /* brute force method */
	    year[0] = myptr[0];
	    year[1] = myptr[1];
	    year[2] = myptr[2];
	    year[3] = myptr[3];
		year[4] = 0x0;

		month[0]= myptr[4];
		month[1]= myptr[5];
		month[2] = 0x0;

		day[0] = myptr[6];
		day[1] = myptr[7];
		day[2] = 0x0;

		hour[0] = myptr[8];
		hour[1] = myptr[9];
		hour[2] = 0x0;

		minutes[0] = myptr[10];
		minutes[1] = myptr[11];
		minutes[2] = 0x0;

		myptr[total_length-1]=0x0;	/*get rid of the Z*/

		seconds = g_malloc(strlen(&myptr[12])+1);
		strcpy (seconds, &myptr[12]);

   	    sprintf(display_text,"%s%s %s-%s-%s %s:%s:%s ",display_string, label_text,
   	    					year , month, day, hour, minutes, seconds );
		g_free(seconds);

	  }

	asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, (guint)asn1->offset - init_offset,display_text);
	g_free(myptr);
	g_free(display_text);
	return(0);
}

/****************************************************************************************/
/* function to handle MMS_BTIME DATA													*/
static void
convert_to_MMSBTIME(ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guint32 total_length;
guint32 saved_length;
guint32 num_msec;
guint32 num_days=0;
guchar display_text[100];
guint32 num_day_seconds;
struct tm *t;
guint32 saved_days;
guint32 num_rem_msec;
gdouble msec_interval;

  total_length = *total_len;
  
  asn1_uint32_value_decode (asn1, 4, &num_msec);
  if (total_length == 6)
    asn1_uint32_value_decode (asn1, 2, &num_days);

  saved_days=num_days;
  num_days = ( num_days+ DAYS_FROM_1970_TO_1984) * SECONDS_PER_DAY;	/*convert days to seconds	*/
 
  num_day_seconds = num_msec/1000;				/*convert msec into whole seconds	*/
  num_days += num_day_seconds;					/*add additional seconds			*/

 
  num_rem_msec = num_msec - (num_day_seconds * 1000);	/*get the number of msec that will be remaining */
  msec_interval = (gdouble)num_rem_msec/1000.0;

 
 saved_length = (asn1->offset-init_offset);
 t = gmtime((const time_t *)&num_days);

 if(saved_days)
    sprintf(display_text,"%s%s %04d-%02d-%02d %02d:%02d:%02.3f (days=%ld msec= %ld)",display_string,
					label_text,1900+t->tm_year,t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min, (gdouble)t->tm_sec +  msec_interval,
					saved_days, num_msec);
 else
    sprintf(display_text,"%s%s %02d:%02d:%02.3f (msec= %ld)",display_string, label_text,
	 t->tm_hour, t->tm_min, (gdouble)t->tm_sec +  msec_interval, num_msec);


 asn1h_proto_tree_add_text (subtree, hf, tvb,  init_offset , saved_length, display_text);
 
 }



/*this routine can be used to skip over any field in their entirety			*/
static void
skip_field(ASN1_SCK *asn1, guchar *display_string, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guchar *myoctets;
guchar *display_text;

		 display_text = g_malloc (strlen(display_string) + strlen(label_text) + 40);
         sprintf(display_text,"%s%s",display_string,label_text);

  	     asn1_string_value_decode (asn1, *total_length, &myoctets);  /*allocate a buffer for the worst case size*/
         asn1h_proto_tree_add_text (subtree, hf, tvb,  init_offset1 , *total_length, display_text);		 
		 g_free(myoctets);
		 g_free(display_text);
}

int convert_to_bitstring_enum (ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text, ENUM_XLAT *xlat)
{
guint32 total_length = *total_len;
guint32 num_unused_bits;
guint32 num_bits_to_display;
guchar *mms_bitstring_data;
char *xlat_string=NULL;


guchar mask = 0x80;
guint32 i;
guint32 byte_offset=0;
guchar data_to_be_masked;
guint32 saved_length; 
guchar mydisplaystring[300];
guint j;


 /*get the length  */
 /*ret = asn1_length_decode (asn1, &def, &total_length); */

  saved_length = (asn1->offset)-init_offset + total_length; 
 
 /*get the number of unused bits */
 asn1_uint32_value_decode (asn1, 1, &num_unused_bits);

 /*now get the bitstring data */
 --total_length;
  num_bits_to_display = (total_length * 8) - num_unused_bits;
  asn1_string_value_decode (asn1, total_length, &mms_bitstring_data);/*allocate a buffer for the worst case size*/

  /*print out the header for the bitstring			*/
  sprintf(mydisplaystring,"%s%s ",display_string,label_text);
  asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_length ,mydisplaystring); 
  
  j = 0;
  for (i=0;i<num_bits_to_display; ++i)
	{
      data_to_be_masked = *(mms_bitstring_data+byte_offset);
	  if (data_to_be_masked & mask)  /*then have a bit */
	  {
	    find_enum_string_value(i,xlat,&xlat_string);
  	    sprintf(mydisplaystring,"%s%s",display_string,xlat_string);
  	    asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, (guint)asn1->offset - init_offset,
  	    			mydisplaystring);
	  }

	  mask = mask >> 1;
	  if (mask == 0x0)
	  {
		  mask = 0x80;
		  ++byte_offset;
	  }

	}


 g_free(mms_bitstring_data);
 return(0);
}

static void
convert_to_unicode (ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guint32 total_length=*total_len;
guchar *myptr;
guchar *display_text;


	asn1_string_value_decode (asn1, total_length, &myptr);

	display_text = g_malloc((total_length*5)+40+strlen(display_string)+strlen(label_text));
 	sprintf(display_text,"%s%s%S",display_string, label_text , myptr);
	asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, (guint)asn1->offset - init_offset,display_text);
	g_free(myptr);
	g_free(display_text);
}


static void
convert_to_vstring (ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
	guint32 total_length=*total_len;
	guchar *myptr;
	guchar *display_text;
	guint32 start_offset_vstring = asn1->offset;

	asn1_string_value_decode (asn1, total_length, &myptr);

	display_text = g_malloc(total_length+40+strlen(display_string)+strlen(label_text));
 	sprintf(display_text,"%s%s%s",display_string, label_text , myptr);

	/* only mark the real visible string for display filtering ! */
	if (display_nodes_as_subtrees)
		asn1h_proto_tree_add_text (subtree, hf, tvb, start_offset_vstring, total_length,myptr);
	else
		asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, (guint)asn1->offset - init_offset,display_text);

	g_free(myptr);
	g_free(display_text);
}


static void
calc_string(guint row_offset, guchar *mydisplaystring, guint num_of_bits, guchar *display_string)
{
 if(num_of_bits!=1)
	{
	sprintf(mydisplaystring,"%s    BIT PAIRS %04u - %04u: ",display_string, 
			(guint)(row_offset*16/num_of_bits),(guint)( (( (row_offset+1)*16)/num_of_bits)-1));
	}
 else
    sprintf(mydisplaystring,"%s    BITS %04u - %04u: ",display_string, 
			row_offset*16, (row_offset*16)+15);
}



static void
convert_to_bitstring (ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text, guint num_of_bits)
{
guint32 total_length = *total_len;
guint32 num_unused_bits;
guint32 num_bits_to_display;
guchar *mms_bitstring_data;


guchar mask = 0x80;
guint32 i;
guint32 byte_offset=0;
guchar data_to_be_masked;
guint32 saved_length; 
guchar mydisplaystring[300];
guint row_offset;
guint j;


 /*get the length  */
 /*ret = asn1_length_decode (asn1, &def, &total_length); */

  saved_length = (asn1->offset)-init_offset + total_length; 
 
 /*get the number of unused bits */
 asn1_uint32_value_decode (asn1, 1, &num_unused_bits);

 /*now get the bitstring data */
 --total_length;
  num_bits_to_display = (total_length * 8) - num_unused_bits;
  asn1_string_value_decode (asn1, total_length, &mms_bitstring_data);/*allocate a buffer for the worst case size*/

  /*print out the header for the bitstring			*/
  sprintf(mydisplaystring,"%s%s ",display_string,label_text);
  asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_length ,mydisplaystring); 
  

  row_offset = 0;
  j=0;
  calc_string(row_offset,mydisplaystring, num_of_bits, display_string);
  for (i=1;i<=num_bits_to_display; ++i)
	{


      data_to_be_masked = *(mms_bitstring_data+byte_offset);

	  if (data_to_be_masked & mask)  /*then have a bit */
		  sprintf(mydisplaystring + strlen(mydisplaystring),"1");
	  else
		  sprintf(mydisplaystring + strlen(mydisplaystring),"0");

	  if(!(i%num_of_bits) )
		  sprintf(mydisplaystring + strlen(mydisplaystring)," ");

	  if (j==15)
	  {
		asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_length ,mydisplaystring);

		j=0;
		++row_offset;
		calc_string(row_offset,mydisplaystring, num_of_bits, display_string);

	  }
	  else
		  ++j;

	  mask = mask >> 1;
	  if (mask == 0x0)
	  {
		  mask = 0x80;
		  ++byte_offset;
	  }

	}

 if(j!=0)
	 asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_length ,mydisplaystring);


 g_free(mms_bitstring_data);
}

static void
call_external_decode(ASN1_SCK *asn1, guchar *display_string, guint32 tag, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text, void *decode_ptr)
{
guchar *myoctets;
guchar *display_text;
specDecodeFunc *fptr = (specDecodeFunc *)decode_ptr;


         display_text = g_malloc(strlen(display_string) + strlen(label_text) + 41 );
  	     asn1_string_value_decode (asn1, *total_length, &myoctets);  /*allocate a buffer for the worst case size*/
		 sprintf(display_text,"%s%s {",display_string, label_text);
		 asn1h_proto_tree_add_text (subtree, hf,tvb,init_offset1,*total_length,display_text);
		 if (fptr !=NULL)
		    fptr (*total_length, myoctets, display_string, init_offset1, subtree, tvb);
		 sprintf(display_text,"%s}", display_string);
		 asn1h_proto_tree_add_text (subtree, hf,tvb,init_offset1,*total_length,display_text);
		 g_free(myoctets);
		 g_free (display_text);
}

/*function to convert and display octetstrings */
static void
convert_to_octetstring(ASN1_SCK *asn1, guchar *display_string, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guchar *myoctets;
guchar *display_text;
guint i;


         display_text = g_malloc( 40 + (*total_length * 3) + strlen(display_string)+strlen(label_text)); 
  	     asn1_string_value_decode (asn1, *total_length, &myoctets);  /*allocate a buffer for the worst case size*/
		 sprintf(display_text,"%s%s",display_string,label_text);
		 for(i=0; i< *total_length; ++i)
			sprintf(display_text+strlen(display_text),"%02x ",*(myoctets+i));
		 asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset1, (guint)asn1->offset - init_offset1,display_text);
		 g_free(myoctets);
		 g_free (display_text);
}
/*function to convert and display unsigned integers */
static void
convert_to_objid(ASN1_SCK *asn1, guchar *display_string, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guchar *display_text;
guint num_ids;
subid_t *oid;
guint i;


         display_text = g_malloc( 64  + strlen(display_string)+strlen(label_text));
		 asn1_oid_value_decode (asn1, *total_length, &oid, &num_ids);
		 display_text = g_malloc(40 + num_ids*6);
		 sprintf(display_text,"%s%s",display_string,label_text);
		 for (i=0;i<num_ids; ++i)
		 {
			 sprintf(display_text+strlen(display_text),"%u ",oid[i]);
		 }

		 asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset1, (guint)asn1->offset - init_offset1,display_text);
		 g_free (display_text);
		 g_free(oid);
}



/*function to convert and display unsigned integers */
static void
convert_to_unsigned_integer(ASN1_SCK *asn1, guchar *display_string, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guchar *display_text;
guint32 myint;


         display_text = g_malloc( 40  + strlen(display_string)+strlen(label_text));
		 asn1_uint32_value_decode (asn1, *total_length, &myint);
		 sprintf(display_text,"%s%s",display_string,label_text);
		 asn1h_proto_tree_add_text_uint(subtree, hf, tvb, init_offset1, (guint)asn1->offset - init_offset1,display_text,myint);
		 g_free (display_text);
}


/*function to convert and display signed integers */
static void
convert_to_integer(ASN1_SCK *asn1, guchar *display_string, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text,
					 ENUM_XLAT *xlat, gboolean display_type_length)
{
guchar *display_text;
gint32 myint;
char *xlat_string=NULL;
gboolean is_variable_length = FALSE;

         display_text = g_malloc( 60  + strlen(display_string)+strlen(label_text));
		 asn1_int32_value_decode (asn1, *total_length, &myint);

		 if(xlat !=NULL)
		 {
		  find_enum_string_value(myint,xlat,&xlat_string);
		  sprintf(display_text,"%s%s%s",display_string,label_text,xlat_string);
		 }
		 else
		 {
		 if (display_type_length ==FALSE)
		   sprintf(display_text,"%s%s",display_string,label_text);
		 else
		   { /*need to determine if fixed or variable length */
			 switch (*total_length)
			 {
			 case 1:			/*one octet value */
			   if (myint & 0x80)
			     is_variable_length = TRUE;
			   myint = myint & 0x7f;
			   break;

			 case 2:			/*two or three octet value*/
			 case 3:
			   if (myint & 0x8000)
			     is_variable_length = TRUE;
			   myint = myint & 0x7fff;
			   break;

			 default:
			   break;
 			 }

		     if (is_variable_length==FALSE)
		   		sprintf(display_text,"%s%s (fixed length)",display_string,label_text);
			 else
		   		sprintf(display_text,"%s%s (variable length)",display_string,label_text);

		   }
		 }
		 asn1h_proto_tree_add_text_int (subtree, hf, tvb, init_offset1, (guint)asn1->offset - init_offset1,display_text,myint);
		 g_free (display_text);
}


/*function to convert and display octetstrings */
static void
convert_to_BOOLEAN(ASN1_SCK *asn1, guchar *display_string, guint32 *total_length, guint32 init_offset1, proto_item *subtree, int hf, tvbuff_t *tvb, guchar *label_text)
{
guchar *myoctets;
guchar *display_text;


         display_text = g_malloc( 40 + (*total_length * 3) + strlen(display_string) +strlen(label_text)); 
  	     asn1_string_value_decode (asn1, 1, &myoctets);  /*allocate a buffer for the worst case size*/
		 if (*myoctets)
		   sprintf(display_text,"%s%s TRUE",display_string,label_text);
		 else
		   sprintf(display_text,"%s%s FALSE",display_string,label_text);

		 asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset1, (guint)asn1->offset - init_offset1,display_text);
		 g_free(myoctets);
		 g_free (display_text);
}



int
asn1h_do_parse (ASN1_SCK *asn1, ASN1_TBL_ENTRY *tbl_array, guchar *display_input, guint32 *total_len, guint32 init_offset1, proto_item *subtree, tvbuff_t *tvb, guint type_of_table,
				gboolean inp_nodes_as_tree)
{
	guint tag;
	guint ret;
	guint internal_len;
	gboolean def;	
	ASN1_TBL_ENTRY *next_tbl=tbl_array;			/*pointer used to end the while loop	*/
	ASN1_TBL_ENTRY *entries= tbl_array;
	guint entry_offset = 0;
	guint init_offset;
	guchar int_display[300];
	guint saved_len;
	guint num_bits_in_pair; 
	guchar display_text[40];
	
	/* we are going to work on total_length		*/
	display_nodes_as_subtrees = inp_nodes_as_tree;			/*set the display mode  */

	while(*total_len!=0)
	{
		int tag_found = FALSE;
		init_offset = asn1->offset;
		asn1h_tag_get (asn1, &tag);
		/*get the tag and the length	*/
		if (type_of_table != ASN1H_TBL_SEQ) /*otherwise will process in order */
			entry_offset = 0;
		while (entries[entry_offset].tag!=ASN1H_END_OF_TABLE)
		{
			int hf = -1;
			int ett = -1;
			if (entries[entry_offset].tag == tag)	/*then we have a match	*/
			  { 
				tag_found = TRUE;			
				ret = asn1h_length_decode (asn1, &def, &internal_len);
				if (!(tag & ASN1H_CONSTRUCT) )			/*then have a primitive  */
				  {
					proto_item *old_subtree = subtree;
					if (display_nodes_as_subtrees)
					{	
						ett = entries[entry_offset].ett ? (*entries[entry_offset].ett) : -1;
						hf = entries[entry_offset].hf ? (*entries[entry_offset].hf) : -1;
					}
					else
					{	
						ett = -1;
						hf = -1;
						/*changed June 15th to allow filtering */
						hf = entries[entry_offset].hf ? (*entries[entry_offset].hf) : -1;
					}
					
					if (ett != -1)
					{
						if (entries[entry_offset].tag_string && internal_len)
						{
							proto_item *p_item;						
							sprintf(int_display,"%s%s",display_input,entries[entry_offset].tag_string);
							p_item = asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_len,
								hf == -1 ? int_display : " "); 
							subtree = proto_item_add_subtree(p_item,ett);								
						}					
					}
					
					switch (entries[entry_offset].conversion_type & ASN1H_DISPLAY_ACTION_MASK)
					{
					case ASN1H_NONASN1_DECODE:
					  {
				      proto_tree *mysubtree=subtree;
						call_external_decode(asn1, display_input, tag, &internal_len, asn1->offset, mysubtree,hf,tvb,
					  		entries[entry_offset].tag_string, (void *)entries[entry_offset].next);
					  }
					break;
			      
					case ASN1H_OCTETSTRING_DISPLAY:
						convert_to_octetstring(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;

					case ASN1H_MMS_TYPE_LENGTH:
						convert_to_integer(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string, NULL,TRUE);
					break;
					
					case ASN1H_INTEGER_ENUM_DISPLAY:
						convert_to_integer(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string, (ENUM_XLAT *)entries[entry_offset].next,FALSE);
					break;
					
					case ASN1H_INTEGER_DISPLAY:
						convert_to_integer(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string, NULL,FALSE);
					break;
					
					
					case ASN1H_OBJID_DISPLAY:
						convert_to_objid(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;

					case ASN1H_BITSTRING_DISPLAY:
						convert_to_bitstring(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string,1);
					break;
					
					case ASN1H_BITSTRING_ENUM_DISPLAY:
						convert_to_bitstring_enum(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string, (ENUM_XLAT *)entries[entry_offset].next);
					break;
					
					case ASN1H_MMS_UNICODE_DISPLAY:
						convert_to_unicode(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					case ASN1H_VSTRING_DISPLAY:
						convert_to_vstring(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					case ASN1H_UNSIGNED_DISPLAY:
						convert_to_unsigned_integer(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					case ASN1H_MMS_BTIME_DISPLAY:
						convert_to_MMSBTIME(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;

					case ASN1H_MMS_UTCTIME_DISPLAY:
						convert_to_MMSUTCTIME(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					
					case ASN1H_GENERALTIME_DISPLAY:
						convert_to_GENERALTIME(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					case ASN1H_BOOLEAN_DISPLAY:
						convert_to_BOOLEAN(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					case ASN1H_NULL_DISPLAY:
						sprintf(display_text,"%s%s",display_input,entries[entry_offset].tag_string);
						asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, (guint)asn1->offset - init_offset,display_text);
					
					break;
														
					case ASN1H_MMS_FLOAT_DISPLAY:
						convert_to_MMSFLOAT(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
							entries[entry_offset].tag_string);
					break;
					
					case ASN1H_DONOT_DISPLAY:
					default:
						if(entries[entry_offset].conversion_type & ASN1H_BITSTRING_PAIR_DISPLAY)
						{
							num_bits_in_pair =	  entries[entry_offset].conversion_type & (~ASN1H_BITSTRING_PAIR_DISPLAY);
							convert_to_bitstring(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
								entries[entry_offset].tag_string,num_bits_in_pair);
						}
						else
							skip_field(asn1, display_input, &internal_len, init_offset, subtree, hf, tvb, 
								entries[entry_offset].tag_string);
					break;
					}  /*end of the switch		*/
					
					if (ett != -1)
						subtree = old_subtree;
					
					*total_len = *total_len - (asn1->offset - init_offset);
					init_offset = asn1->offset;
					if (type_of_table & ASN1H_TBL_SEQ) /*otherwise will process in order */
					    ++entry_offset;
 	  
				  }	/*end of the primitive tag processing if a match	*/
				else	
				  { /*have a matched constructor					*/
				  saved_len = (guint) asn1->offset - init_offset+internal_len;
				  tag_found = TRUE;

                  /* added to allow specialized decodes, even in constructors */
				  if (entries[entry_offset].conversion_type == ASN1H_NONASN1_DECODE)
						call_external_decode(asn1, display_input, tag, &internal_len, asn1->offset, subtree,hf,tvb,
							entries[entry_offset].tag_string, (void *)entries[entry_offset].next);

				  /*now create another indent and pass it on			*/

				  if ( (entries[entry_offset].conversion_type != ASN1H_DO_NOT_PROCESS) &&
				       (entries[entry_offset].conversion_type != ASN1H_NONASN1_DECODE) )
					{
					  /* now we need to print a { in order to indent */
					
				      proto_item *old_subtree = subtree;
					  if (display_nodes_as_subtrees)
					    {	
					     ett = entries[entry_offset].ett ? (*entries[entry_offset].ett) : -1;
					     hf  = entries[entry_offset].hf  ? (*entries[entry_offset].hf) : -1;
					    }
					  else
					    {
					     ett = -1;
						 hf  = -1;
					    }

					  if (ett != -1)
						{
						  if (entries[entry_offset].tag_string && internal_len)
							{
							  /* then we need to display the string prior to indenting	*/
							  proto_item *p_item;
							  sprintf(int_display,"%s%s",display_input,entries[entry_offset].tag_string);
							  p_item = asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_len,
							  		hf == -1 ? int_display : " ");
							  subtree = proto_item_add_subtree(p_item,ett);								
							}
						  else
							{
							  sprintf(int_display,"%s",display_input);
							}						
						}
					  else
						{
						  if (entries[entry_offset].tag_string && internal_len)
							{
							  /* then we need to display the string prior to indenting	*/
							  sprintf(int_display,"%s%s",display_input,entries[entry_offset].tag_string);
							  asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_len,int_display);
							}

						  if (!display_nodes_as_subtrees)
							{
							  sprintf(int_display,"%s{",display_input);
							  asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_len,int_display);
							}
						}

					  sprintf(int_display,"%s  ",display_input);


					  if (entries[entry_offset].next !=NULL)
							next_tbl = entries[entry_offset].next;			/*set up for the next table	*/
					  else
							next_tbl = entries;

					  if (next_tbl && entries[entry_offset].tag)
					  	asn1h_do_parse (asn1, next_tbl, int_display, &internal_len, asn1->offset, subtree, tvb, entries[entry_offset].conversion_type,
										inp_nodes_as_tree);

					  if (type_of_table & ASN1H_TBL_SEQ)
							++entry_offset; 

					  /* once returned, need to end the indent	*/
						
					  if (ett != -1)
						{
							subtree = old_subtree;
						}
					  else
						{
						  if (!display_nodes_as_subtrees)
							{						
							  sprintf(int_display,"%s}",display_input);
							  asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_len,int_display);
							}
						}

					  *total_len = *total_len - saved_len;						

					  if (entries[entry_offset].tag_string && internal_len)
					    {
						  /*then we need to display the string prior to indenting	*/
						  hf = entries[entry_offset].hf ? (*entries[entry_offset].hf) : -1;					
						  sprintf(int_display,"%s%s",display_input,entries[entry_offset].tag_string);
						  asn1h_proto_tree_add_text (subtree, hf, tvb, init_offset, saved_len,int_display);
					    }
				    }	/*end of DO NOT PROCESS	*/
				  else
					*total_len = *total_len - saved_len;

				  }		/*end of matched constructor processing	*/
			    break;	/*break the while loop that is searching based upon entry offset */
			    }	    /*end of matched processing				*/
						/*need to go get the next tag if there is length still remaining	*/
		    else
			  ++entry_offset;
		  } 	/*end of while loop that is based upon entry_offset 		*/

		if (!tag_found) 
		  {
/*			ret = asn1h_length_decode (asn1, &def, &internal_len);  */
			sprintf(int_display,"%s: TAG 0x%x (%u) not found in table may be another protocol !"
				,display_input,tag,tag);
			asn1h_proto_tree_add_text (subtree, -1, tvb, init_offset, saved_len, int_display);
		  }

			
		if (entries[entry_offset].tag==ASN1H_END_OF_TABLE)
		  {	
			*total_len = 0;
		  }
	}
	return(0);
}