/* packet-isopres.c
*
* Routine to dissect connectionless ISO 8823-1 OSI Presentation Protocol packets
*
* $Id: packet-isopres.c,v 1.0 2004/012/12  Exp $
*
* Herbert Falk <herb@sisconet.com>
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>

#include <stdio.h>
#include <string.h>
#include <epan/conversation.h>
#include <epan/tvbuff.h>
#include "prefs.h"
#include "reassemble.h"
#include "tap.h"
#include "packet-frame.h"

#include "packet-clisopres.h"
#include "asn1.h"
#include "asn1_helper.h"
#include "packet-frame.h"

#include <epan/strutil.h>



#define NOET NULL
#define NOHF NULL

/* find the dissector for data */
static dissector_handle_t adata_handle;
static dissector_handle_t acse_handle;
static dissector_handle_t mms_handle;


static int proto_isopres          = -1;
static int proto_clisopres          = -1;
static gint ett_isopres           = -1;


#define ISO_PRES_CO				0			/*Connection Oriented Presentation	*/
#define ISO_PRES_CL				1			/*Connectionless Presentation	*/
static int hf_isopres_type        = -1;

#define CL_PRES_SEQ		0x30		/*sequence code for Connectionless Presentation */
#define CO_PRES_PDATA	(ASN1H_APP | ASN1H_CONSTRUCT |1)		/*PDATA			*/

#define CL_PROTO_VER	(ASN1H_CTX | 0 )
#define CALL_PSEL		(ASN1H_CTX | 1 )
#define CALLED_PSEL		(ASN1H_CTX | 2 )
#define P_CONTEXT_LIST	(ASN1H_CTX | ASN1H_CONSTRUCT| 4)
#define P_FULLY_ENCODED_DATA (ASN1H_APP | ASN1H_CONSTRUCT | 1)


#define PROTO_STRING_PRES	"ISO 8823 Presentation"
#define PROTO_STRING_CLPRES  "ISO 8823 CL Presentation"
#define ASN1H_SEQ_TAG	(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ)
#define ASN1H_INTEGER_ID (ASN1H_UNI | ASN1_INT)		/*ASN1_INT is defined in ASN1.h*/
#define ASN1H_OBJECT_ID (ASN1H_UNI | ASN1_OJI)		/*ASN1_OJI is defined in ASN1.h*/

 proto_item *data_item;
static guchar PSEL_PROTO [] = "PROTOCOL VERSION: ";
static guchar CALL_PSEL_TEXT [] = "Calling PSEL: ";
static guchar CALLED_PSEL_TEXT [] = "Called PSEL: ";
static guchar PCTXT_ID[] = "PCTXT ID: ";
static guchar AS_ID[] = "Abstract Syntax: ";
static guchar TS_ID[] = "Transfer Syntax: ";
static guchar TS_NAME[] = "Transfer ID: ";
static gboolean desegment_pres = TRUE;


static gboolean dissect_isopres_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree);


/************************* Beginning of Connectionless Presentation Tables	*/

extern ASN1_TBL_ENTRY pres_ts_seq[2];

static ASN1_TBL_ENTRY end_of_pres[]={
	{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY pdv_list_start[] ={
	{ASN1H_OBJECT_ID,NOET, NOHF, TS_ID, ASN1H_OBJID_DISPLAY, pdv_list_start},	/*Transfer syntax name */
	{ASN1H_INTEGER_ID,NOET, NOHF, PCTXT_ID,ASN1H_INTEGER_DISPLAY,pdv_list_start},
	{ASN1H_SEQ_TAG,NOET, NOHF, NULL, ASN1H_SKIP_TAG_AND_LENGTH, pdv_list_start},
	{(ASN1H_CTX | ASN1H_CONSTRUCT),NOET, NOHF, "Single ASN1 Type", ASN1H_DO_NOT_PROCESS,end_of_pres},
	{(ASN1H_CTX | 1),NOET, NOHF, "Octet Aligned",ASN1H_DO_NOT_PROCESS,end_of_pres},
	{(ASN1H_CTX | 2),NOET, NOHF, "Arbitray",ASN1H_DO_NOT_PROCESS, end_of_pres},
	{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
};





static ASN1_TBL_ENTRY fully_encoded_data [] = {
	{ASN1H_SEQ_TAG,NOET, NOHF, "PDV LIST", ASN1H_SKIP_TAG_AND_LENGTH, pdv_list_start},
	{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY pres_ctxt_list[] ={
	{ASN1H_INTEGER_ID,NOET, NOHF, PCTXT_ID,ASN1H_INTEGER_DISPLAY,pres_ctxt_list},
	{ASN1H_OBJECT_ID,NOET, NOHF, AS_ID, ASN1H_OBJID_DISPLAY, pres_ctxt_list},
	{ASN1H_SEQ_TAG,NOET, NOHF, "TS Names", ASN1H_INDENT, pres_ts_seq},
	{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
	};


static ASN1_TBL_ENTRY pres_ctxt_list_start[]={
		{ASN1H_SEQ_TAG,NOET, NOHF,  NULL, ASN1H_INDENT, pres_ctxt_list },
		{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
		};

static ASN1_TBL_ENTRY pres_ts_seq[]={
	{ASN1H_OBJECT_ID,NOET, NOHF,  TS_ID, ASN1H_OBJID_DISPLAY, pres_ctxt_list_start},
	{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY isopres_main[]={
		{CL_PROTO_VER, NOET, NOHF, PSEL_PROTO, ASN1H_INTEGER_DISPLAY, isopres_main },
		{CALL_PSEL, NOET, NOHF, CALL_PSEL_TEXT, ASN1H_OCTETSTRING_DISPLAY, isopres_main },
		{CALLED_PSEL, NOET, NOHF, CALLED_PSEL_TEXT, ASN1H_OCTETSTRING_DISPLAY, isopres_main },
	    {P_FULLY_ENCODED_DATA,NOET, NOHF, "FULL CODED DATA", ASN1H_INDENT, fully_encoded_data},
		{P_CONTEXT_LIST,NOET, NOHF, "P_CTXT_LIST:", ASN1H_INDENT, pres_ctxt_list_start}, 	/*don't display the p_context list */
		{ASN1H_END_OF_TABLE	,NOET, NOHF, NULL,NULL,NULL}
		};		 

/************************* End of connectionless presentation tables 			*/


/************************ Beginning of Connection Oriented Presentation Tables	*/
static ASN1_TBL_ENTRY pdata_list_start[] ={
	{ASN1H_SEQ_TAG,NOET,NOHF,NULL, ASN1H_SKIP_TAG_AND_LENGTH, pdata_list_start},
	{ASN1H_INTEGER_ID,NOET, NOHF,PCTXT_ID,ASN1H_INTEGER_DISPLAY,pdv_list_start},
		{ASN1H_END_OF_TABLE	,NOET,NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY iso_co_pres_main[]={
		{CO_PRES_PDATA,NOET, NOHF, "PDATA", ASN1H_TBL_SEQ , pdata_list_start},
		{ASN1H_END_OF_TABLE	,NOET,NOHF,NULL,NULL,NULL}
		};		 


static int
get_item_len(tvbuff_t *tvb, int offset, int *len_len)
{
	guint16 len;

	len = tvb_get_guint8(tvb, offset);
	if(len == 0x82)
	{
		len = tvb_get_ntohs(tvb, offset+1);
		*len_len = 3;
	}
	else
		*len_len = 1;
	return len;
}

static void
dissect_cl_presentation(ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, tvbuff_t *tvb)
{
guint tag;
guint total_length;
gboolean not_end = FALSE;
guchar *mysel;
guint i;
guchar display_text[32];
gint ret;
gboolean def;
guint init_offset1;

   asn1h_do_parse (asn1, isopres_main, "", total_len, asn1->offset, subtree, tvb, ASN1H_TBL_SET,FALSE);
}


static void
dissect_co_presentation(ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, tvbuff_t *tvb)
{
guint tag;
guint total_length;
gboolean not_end = FALSE;
guchar *mysel;
guint i;
guchar display_text[32];
gint ret;
gboolean def;
guint init_offset1;

   asn1h_do_parse (asn1, iso_co_pres_main, "", total_len, asn1->offset, subtree, tvb, ASN1H_TBL_SEQ,FALSE);
}

static void
dissect_isopres(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{

	int offset = 0;
	guint8 type;
	proto_item *ti;
    proto_tree  *isopres_tree;
	guint tag;
	int ret;
	gboolean def;
	ASN1_SCK asn1;
	guint total_length;
	tvbuff_t *next_tvb;
	guint saved_length;
	guint8 *cont_buffer;




	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "PRES");
  	if (check_col(pinfo->cinfo, COL_INFO))
  		col_clear(pinfo->cinfo, COL_INFO);

    if (!tree)
	{
      asn1_open(&asn1, tvb, offset);
      ret = asn1_tag_decode (&asn1, &tag);
	  if(tag ==CL_PRES_SEQ)
	  {
	    if(adata_handle)
	      call_dissector(adata_handle, next_tvb, pinfo, tree);
	  }
	  else
	  {
		if (tag == 0x61)
		{
		  if(mms_handle)
		      call_dissector(mms_handle, next_tvb, pinfo, tree);
		}
	  }

	}

	if (tree) {
	  /*open the ASN1 tree, offset is tvb->offset */
      asn1_open(&asn1, tvb, offset);

    	ret = asn1_tag_decode (&asn1, &tag);


	  if (tag ==CL_PRES_SEQ)
	    {
		ret = asn1_length_decode (&asn1, &def, &total_length);
		ti = proto_tree_add_item(tree, proto_clisopres, tvb, 0, -1, FALSE);
  		data_item = proto_item_add_subtree(ti, ett_isopres); 

		 dissect_cl_presentation (&asn1, tree, &total_length, offset, data_item, tvb);
		 /*call to parse AUNIT-DATA */

	    proto_item_set_end(ti, tvb, offset);

		if (tvb_reported_length_remaining(tvb, asn1.offset) > 0) 
			next_tvb = tvb_new_subset(tvb, asn1.offset, -1, -1);

		 if(adata_handle)
		     call_dissector(adata_handle, next_tvb, pinfo, tree);



		} 
	  else
	    {
		if(tvb->length < tvb->reported_length)
			tvb->length = tvb->reported_length;
#if 0
		asn1.offset -=1;			/*move back one byte		*/
#endif
		switch (tag)
		  {
		  case 0x61:

	     	ti = proto_tree_add_item(tree, proto_isopres, tvb, 0, -1, FALSE);
 
	 		data_item = proto_item_add_subtree(ti, ett_isopres); 
			total_length = tvb->reported_length;
			
			saved_length = total_length;

			dissect_co_presentation (&asn1, tree, &total_length, offset, data_item, tvb);
	
/*			ret = asn1h_tag_get (&asn1, &tag);
	        proto_item_set_end(ti, tvb, offset); */

		    ret = asn1h_length_decode (&asn1, &def, &total_length);

			if (tvb_reported_length_remaining(tvb, asn1.offset) > 0) 
			  next_tvb = tvb_new_subset(tvb, asn1.offset, -1, total_length);


			if(mms_handle)
		      call_dissector(mms_handle, next_tvb, pinfo, tree);
			break;

		  default:
		    break;
		  }
        }
	}
}

void
proto_register_isopres(void)
{
	static hf_register_info hf[] =
	{

		{
			&hf_isopres_type,
			{
				"SPDU Type",
				"ses.type",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x0,
				"", HFILL
			}
		},
	};

	static gint *ett[] =
	{
		&ett_isopres,
	};
	module_t *pres_module;


	proto_clisopres = proto_register_protocol(PROTO_STRING_CLPRES, "CLPR", "prcl");
	proto_register_field_array(proto_clisopres, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	pres_module = prefs_register_protocol(proto_clisopres, NULL);
/*
	prefs_register_bool_preference(ses_module, "desegment",
	    "Desegment all session packets ",
	    "Whether the session dissector should desegment all messages spanning multiple SES segments",
	    &ses_desegment);  */

	/*
	 * Register the dissector by name, so other dissectors can
	 * grab it by name rather than just referring to it directly
	 * (you can't refer to it directly from a plugin dissector
	 * on Windows without stuffing it into the Big Transfer Vector).
	 */
/*	register_dissector("pres", dissect_isopres, proto_isopres); */
	register_dissector("prcl", dissect_isopres_heur, proto_isopres);
}


static guint
get_pres_pdu_len(tvbuff_t *tvb, int offset)
{
	ASN1_SCK asn1;
	guint tag;
	int ret;
	guint total_length;
	gboolean def;


	/*  OK,let's check SPDU length  */
	/*  get length of SPDU */
    asn1_open(&asn1, tvb, offset);
   	ret = asn1_tag_decode (&asn1, &tag);
	ret = asn1_length_decode (&asn1, &def, &total_length);
	return(total_length);
}


/*static gboolean*/
static gboolean
dissect_isopres_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree)
{
	/* must check that this really is a ses packet */
	int offset = 0;
	guint8 type;
	int len_len;
	guint16 len;
	ASN1_SCK asn1;
	guint tag;
	int ret;
	guint total_length;
	gboolean def;



	/* can we regognize session PDU ? Return FALSE if  not */
	/*   get SPDU type */
	type = tvb_get_guint8(tvb, offset);
	/* check SPDU type */

	/*  OK,let's check SPDU length  */
	/*  get length of SPDU */
    asn1_open(&asn1, tvb, offset);
   	ret = asn1_tag_decode (&asn1, &tag);
	ret = asn1_length_decode (&asn1, &def, &total_length);



	/* do we have enough bytes ? */

	if (!tvb_bytes_exist(tvb, 0, total_length))
	{
		desegment_pres = TRUE;
		pinfo->can_desegment = 2;
        tcp_dissect_pdus(tvb, pinfo, parent_tree,
                   desegment_pres,        /* desegment or not   */
                   8,                     /* magic + length */
                   get_pres_pdu_len,      /* use first 4, calc data len */
                   dissect_isopres); /* the naive dissector */
		return(0);
	}
	else
	{
	  	dissect_isopres(tvb, pinfo, parent_tree);
		return(1);
	}
#if 0
	  	dissect_isopres(tvb, pinfo, parent_tree);
		return(1);
#endif
}

void
proto_reg_handoff_isopress(void)
{
guchar myname[32];



	/* define sub dissector */
	strcpy(myname, "adata");
	adata_handle = find_dissector(myname);


}
