/* packet-iecgsse.c
* Routines for IEC 61850 IECGSSE dissection
* 2003, Herbert Falk <herb@sisconet.com>
*
* Herbert Falk <herb@sisconet.com>
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "etypes.h"				/*for the Ethertype IDs, may no longer be needed*/
#include "asn1.h"				/*for the ASN1 decoder functions */
#include "asn1_helper.h"



static int proto_iecgsse          = -1;
static gint ett_iecgsse           = -1;
static gint ett_iecgsse_appid           = -1;
static gint ett_iecgsse_state           = -1;
static gint hf_iecgsse_type		  = -1;
static gint hf_iecgsse_appid	  = -1;
static gint hf_iecgsse_state 	  = -1;
static proto_item *data_item;
static int displayed_header=FALSE;

#define NOETT NULL
#define NOHF  NULL


#define PROTO_STRING_GSSE	"IEC 61850 GSSE"
#define CL_GSSE_SEQ		(ASN1H_CTX | ASN1H_CONSTRUCT | 3)		/*sequence code for GSSE */

#define ASN1H_SEQ_TAG	(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ)
#define ASN1H_INTEGER_ID (ASN1H_UNI | ASN1_INT)		/*ASN1_INT is defined in ASN1.h*/
#define ASN1H_OBJECT_ID (ASN1H_UNI | ASN1_OJI)		/*ASN1_OJI is defined in ASN1.h*/


static ASN1_TBL_ENTRY gsse_struct[]={
		{(ASN1H_CTX | 10)	,NOETT, &hf_iecgsse_appid, "", ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 12)   ,NOETT, NOHF, "Event Time: ", ASN1H_MMS_BTIME_DISPLAY, NULL},
		{(ASN1H_CTX | 6)  , NOETT, NOHF,"Sequence Number: ", ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 6)  , NOETT, &hf_iecgsse_state,"", ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 6)  , NOETT, NOHF,"HoldTime (msec): ", ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 6)  ,NOETT, NOHF, "usecs: ", ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 6)  ,NOETT, NOHF, "PhsID: ", ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 4)  ,NOETT, NOHF, "DNA: ", ASN1H_BITSTRING_PAIR_DISPLAY | 2 , NULL},
		{(ASN1H_CTX | 4)  ,NOETT, NOHF, "UserST: ", ASN1H_BITSTRING_PAIR_DISPLAY | 2, NULL},
 		{ASN1H_END_OF_TABLE,NOETT, NOHF,NULL,NULL,NULL}
};




static ASN1_TBL_ENTRY data_struct[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT |2)	,NOETT, NOHF,NULL,ASN1H_TBL_SEQ,gsse_struct},
		{ASN1H_END_OF_TABLE	,NOETT, NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY obj_name[]={
		{ (ASN1H_CTX | 0), NOETT, NOHF,"NAME: ", ASN1H_VSTRING_DISPLAY, NULL},
        { (ASN1H_CTX | ASN1H_CONSTRUCT | 0),NOETT, NOHF, "GSSE DATA", NULL, data_struct},
		{ASN1H_END_OF_TABLE	,NOETT, NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY obj_outer[]={
        { (ASN1H_CTX | ASN1H_CONSTRUCT | 0), NOETT, NOHF,NULL, NULL, obj_name},
		{ASN1H_END_OF_TABLE	,NOETT, NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY var_spec[]={
		{ ASN1H_SEQ_TAG ,NOETT, NOHF, "OBJECTName", NULL, obj_outer},
		{ASN1H_END_OF_TABLE	,NOETT, NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY info_rpt[]={
        { (ASN1H_CTX | ASN1H_CONSTRUCT | 0), NOETT, NOHF,"VAR SPEC", NULL, var_spec},		/*is the information report */
        { (ASN1H_CTX | ASN1H_CONSTRUCT | 0), NOETT, NOHF,"DATA_RESULTS", ASN1H_TBL_SEQ, data_struct},		/*is the information report */
		{ASN1H_END_OF_TABLE	,NOETT, NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY iecgsse_main[]={
        { (ASN1H_CTX | ASN1H_CONSTRUCT | 0),NOETT, NOHF, NULL,ASN1H_TBL_SEQ, info_rpt},		/*the first AO (sequence for unconfirmed */
		{ASN1H_END_OF_TABLE	,NOETT, NOHF,NULL,NULL,NULL}
		};		 

static void
dissect_iecgsse(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset = 0;
	proto_item *ti;
	guint tag;
	int ret;
	gboolean def;
	ASN1_SCK asn1;
	guint total_length;					 


	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "GSSE");
  	if (check_col(pinfo->cinfo, COL_INFO))
  		col_clear(pinfo->cinfo, COL_INFO);

	if (check_col(pinfo->cinfo, COL_INFO)) 
	{
	  if (displayed_header == FALSE)
	  {
		col_set_str(pinfo->cinfo, COL_INFO, "Decoder Courtesy of Herbert Falk of SISCO");
		displayed_header = TRUE;
	  }
 	}

	if (tree) 
	  {
	  /*open the ASN1 tree, offset is tvb->offset */
      asn1_open(&asn1, tvb, offset);

   	  ret = asn1h_tag_get (&asn1, &tag);

      ret = asn1h_length_decode (&asn1, &def, &total_length);
	  if (tag ==CL_GSSE_SEQ)
	    {
		ti = proto_tree_add_item(tree, proto_iecgsse, tvb, 0, -1, FALSE);
  		data_item = proto_item_add_subtree(ti, ett_iecgsse); 

        asn1h_do_parse (&asn1, iecgsse_main, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET, FALSE);
		} 
	  }
}


void
proto_register_iecgsse(void)
{
	static hf_register_info hf[] =
	{

		{  	&hf_iecgsse_type,
			{"SPDU Type", "ses.type", FT_UINT8,	BASE_DEC,NULL,0x0,"", HFILL}
		},
		{ 	&hf_iecgsse_appid,
			{ "  GSID*", "gsse.gsid", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},
		{ 	&hf_iecgsse_state,
			{ "  StateNumber*", "gsse.state", FT_UINT32, BASE_DEC, NULL, 0x0,"", HFILL}
		},
	};

	static gint *ett[] =
	{
		&ett_iecgsse,
		&ett_iecgsse_appid,
		&ett_iecgsse_state,
	};


	proto_iecgsse = proto_register_protocol(PROTO_STRING_GSSE, "IECGSSE", "gsse");
	proto_register_field_array(proto_iecgsse, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	register_dissector("gsse", dissect_iecgsse, proto_iecgsse);
}

void
proto_reg_handoff_iecgsse(void)
{

}

